import base64
import mimetypes
import os

from django.conf import settings
from django.http import Http404, HttpResponse
from django.shortcuts import get_object_or_404
from django.views.decorators.cache import cache_control
from django.views.static import serve as django_serve

from database_files.models import File

@cache_control(max_age=86400)
def serve(request, name):
    """
    Retrieves the file from the database.
    """
    f = get_object_or_404(File, name=name)
    f.dump()
    mimetype = mimetypes.guess_type(name)[0] or 'application/octet-stream'
    response = HttpResponse(f.content, content_type=mimetype)
    response['Content-Length'] = f.size
    return response

def serve_mixed(request, *args, **kwargs):
    """
    First attempts to serve the file from the filesystem,
    then tries the database.
    """
    name = kwargs.get('name') or kwargs.get('path')
    document_root = kwargs.get('document_root')
    document_root = document_root or settings.MEDIA_ROOT
    try:
        # First attempt to serve from filesystem.
        return django_serve(request, name, document_root)
    except Http404:
        # Then try serving from database.
        return serve(request, name)
    


import os

from django.conf import settings
from django.core.urlresolvers import reverse

# If true, when file objects are created, they will be automatically copied
# to the local file system for faster serving.
settings.DB_FILES_AUTO_EXPORT_DB_TO_FS = getattr(
    settings,
    'DB_FILES_AUTO_EXPORT_DB_TO_FS',
    True)

def URL_METHOD_1(name):
    """
    Construct file URL based on media URL.
    """
    return os.path.join(settings.MEDIA_URL, name)

def URL_METHOD_2(name):
    """
    Construct file URL based on configured URL pattern.
    """
    return reverse('database_file', kwargs={'name': name})

URL_METHODS = (
    ('URL_METHOD_1', URL_METHOD_1),
    ('URL_METHOD_2', URL_METHOD_2),
)

settings.DATABASE_FILES_URL_METHOD_NAME = getattr(
    settings,
    'DATABASE_FILES_URL_METHOD',
    'URL_METHOD_1')

if callable(settings.DATABASE_FILES_URL_METHOD_NAME):
    method = settings.DATABASE_FILES_URL_METHOD_NAME
else:
    method = dict(URL_METHODS)[settings.DATABASE_FILES_URL_METHOD_NAME]
settings.DATABASE_FILES_URL_METHOD = method

settings.DB_FILES_DEFAULT_ENFORCE_ENCODING = getattr(
    settings, 'DB_FILES_DEFAULT_ENFORCE_ENCODING', True)

settings.DB_FILES_DEFAULT_ENCODING = getattr(
    settings,
    'DB_FILES_DEFAULT_ENCODING',
    'ascii')

settings.DB_FILES_DEFAULT_ERROR_METHOD = getattr(
    settings, 'DB_FILES_DEFAULT_ERROR_METHOD', 'ignore')

settings.DB_FILES_DEFAULT_HASH_FN_TEMPLATE = getattr(
    settings, 'DB_FILES_DEFAULT_HASH_FN_TEMPLATE', '%s.hash')



import os
from optparse import make_option

from django.core.management.base import BaseCommand, CommandError

from database_files.models import File

class Command(BaseCommand):
    option_list = BaseCommand.option_list + (
#        make_option('-w', '--overwrite', action='store_true',
#            dest='overwrite', default=False,
#            help='If given, overwrites any existing files.'),
    )
    help = 'Dumps all files in the database referenced by FileFields ' + \
        'or ImageFields onto the filesystem in the directory specified by ' + \
        'MEDIA_ROOT.'

    def handle(self, *args, **options):
        File.dump_files(verbose=True)
        


from __future__ import print_function

import base64

import six

from django.conf import settings
from django.db import models
from django.utils import timezone

from database_files import utils
from database_files.utils import write_file, is_fresh
from database_files.manager import FileManager

from . import settings as _settings

class File(models.Model):
    
    objects = FileManager()
    
    name = models.CharField(
        max_length=255,
        unique=True,
        blank=False,
        null=False,
        db_index=True)
    
    size = models.PositiveIntegerField(
        db_index=True,
        blank=False,
        null=False)

    _content = models.TextField(db_column='content')
    
    created_datetime = models.DateTimeField(
        db_index=True,
        default=timezone.now,
        verbose_name="Created datetime")
    
    _content_hash = models.CharField(
        db_column='content_hash',
        db_index=True,
        max_length=128,
        blank=True, null=True)
    
    class Meta:
        db_table = 'database_files_file'
    
    def save(self, *args, **kwargs):
        
        # Check for and clear old content hash.
        if self.id:
            old = File.objects.get(id=self.id)
            if old._content != self._content:
                self._content_hash = None
                
        # Recalculate new content hash.
        self.content_hash
        
        return super(File, self).save(*args, **kwargs)
    
    @property
    def content(self):
        c = self._content
        if not isinstance(c, six.binary_type):
            c = c.encode('utf-8')
        return base64.b64decode(c)
    
    @content.setter
    def content(self, v):
        self._content = base64.b64encode(v)
        
    @property
    def content_hash(self):
        if not self._content_hash and self._content:
            self._content_hash = utils.get_text_hash(self.content)
        return self._content_hash
    
    def dump(self, check_hash=False):
        """
        Writes the file content to the filesystem.
        
        If check_hash is true, clears the stored file hash and recalculates.
        """
        if is_fresh(self.name, self._content_hash):
            return
        write_file(
            self.name,
            self.content,
            overwrite=True)
        if check_hash:
            self._content_hash = None
        self.save()
    
    @classmethod
    def dump_files(cls, debug=True, verbose=False):
        """
        Writes all files to the filesystem.
        """
        if debug:
            tmp_debug = settings.DEBUG
            settings.DEBUG = False
        try:
            q = cls.objects.only('id', 'name', '_content_hash')\
                .values_list('id', 'name', '_content_hash')
            total = q.count()
            if verbose:
                print('Checking %i total files...' % (total,))
            i = 0
            for (file_id, name, content_hash) in q.iterator():
                i += 1
                if verbose and not i % 100:
                    print('%i of %i' % (i, total))
                if not is_fresh(name=name, content_hash=content_hash):
                    if verbose:
                        print(('File %i-%s is stale. Writing to local file '
                            'system...') % (file_id, name))
                    f = File.objects.get(id=file_id)
                    write_file(
                        f.name,
                        f.content,
                        overwrite=True)
                    f._content_hash = None
                    f.save()
        finally:
            if debug:
                settings.DEBUG = tmp_debug



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-10-11 18:11
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='File',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(db_index=True, max_length=255, unique=True)),
                ('size', models.PositiveIntegerField(db_index=True)),
                ('_content', models.TextField(db_column='content')),
                ('created_datetime', models.DateTimeField(db_index=True, default=django.utils.timezone.now, verbose_name='Created datetime')),
                ('_content_hash', models.CharField(blank=True, db_column='content_hash', db_index=True, max_length=128, null=True)),
            ],
            options={
                'db_table': 'database_files_file',
            },
        ),
    ]



# -*- coding: utf-8 -*-
import os
import shutil

import six
from six import StringIO

from django.core import files
from django.test import TestCase
from django.core.files.storage import default_storage
from django.core.management import call_command
from django.conf import settings

from database_files.models import File
from database_files.tests.models import Thing
from database_files import utils

DIR = os.path.abspath(os.path.split(__file__)[0])

class DatabaseFilesTestCase(TestCase):
    
    def setUp(self):
        self.media_dir = os.path.join(DIR, 'media/i/special')
        if os.path.isdir(self.media_dir):
            shutil.rmtree(self.media_dir)
        os.makedirs(self.media_dir)
    
    def test_adding_file(self):
        
        # Create default thing storing reference to file
        # in the local media directory.
        test_fqfn = os.path.join(self.media_dir, 'test.txt')
        open(test_fqfn, 'w').write('hello there')
        o1 = o = Thing()
        test_fn = 'i/special/test.txt'
        o.upload = test_fn
        o.save()
        obj_id = o.id
        
        # Confirm thing was saved.
        Thing.objects.update()
        q = Thing.objects.all()
        self.assertEqual(q.count(), 1)
        self.assertEqual(q[0].upload.name, test_fn)
        
        # Confirm the file only exists on the file system
        # and hasn't been loaded into the database.
        q = File.objects.all()
        self.assertEqual(q.count(), 0)
        
        # Verify we can read the contents of thing.
        o = Thing.objects.get(id=obj_id)
        self.assertEqual(o.upload.read(), b"hello there")
        
        # Verify that by attempting to read the file, we've automatically
        # loaded it into the database.
        File.objects.update()
        q = File.objects.all()
        self.assertEqual(q.count(), 1)
        self.assertEqual(q[0].content, b"hello there")
        
        # Load a dynamically created file outside /media.
        test_file = files.temp.NamedTemporaryFile(
            suffix='.txt',
            # Django>=1.10 no longer allows accessing files outside of MEDIA_ROOT...
            #dir=files.temp.gettempdir()
            dir=os.path.join(settings.PROJECT_DIR, 'media'),
        )
        data0 = b'1234567890'
        test_file.write(data0)
        test_file.seek(0)
        t = Thing.objects.create(
            upload=files.File(test_file),
        )
        self.assertEqual(File.objects.count(), 2)
        t = Thing.objects.get(pk=t.pk)
        self.assertEqual(t.upload.file.size, 10)
        self.assertEqual(t.upload.file.name[-4:], '.txt')
        self.assertEqual(t.upload.file.read(), data0)
        t.upload.delete()
        self.assertEqual(File.objects.count(), 1)
        
        # Delete file from local filesystem and re-export it from the database.
        self.assertEqual(os.path.isfile(test_fqfn), True)
        os.remove(test_fqfn)
        self.assertEqual(os.path.isfile(test_fqfn), False)
        o1.upload.read() # This forces the re-export to the filesystem.
        self.assertEqual(os.path.isfile(test_fqfn), True)
        
        # This dumps all files to the filesystem.
        File.dump_files()
        
        # Confirm when delete a file from the database, we also delete it from
        # the filesystem.
        self.assertEqual(default_storage.exists('i/special/test.txt'), True)
        default_storage.delete('i/special/test.txt')
        self.assertEqual(default_storage.exists('i/special/test.txt'), False)
        self.assertEqual(os.path.isfile(test_fqfn), False)

    def test_hash(self):
        verbose = 1
        
        # Create test file.
        image_content = open(os.path.join(DIR, 'fixtures/test_image.png'), 'rb').read()
        fqfn = os.path.join(self.media_dir, 'image.png')
        open(fqfn, 'wb').write(image_content)
        
        # Calculate hash from various sources and confirm they all match.
        expected_hash = '35830221efe45ab0dc3d91ca23c29d2d3c20d00c9afeaa096ab256ec322a7a0b3293f07a01377e31060e65b4e5f6f8fdb4c0e56bc586bba5a7ab3e6d6d97a192' # pylint: disable=C0301
        h = utils.get_text_hash(image_content)
        self.assertEqual(h, expected_hash)
        h = utils.get_file_hash(fqfn)
        self.assertEqual(h, expected_hash)
        h = utils.get_text_hash(open(fqfn, 'rb').read())
        self.assertEqual(h, expected_hash)
#        h = utils.get_text_hash(open(fqfn, 'r').read())#not supported in py3
#        self.assertEqual(h, expected_hash)
        
        # Create test file.
        if six.PY3:
            image_content = six.text_type('aあä')#, encoding='utf-8')
        else:
            image_content = six.text_type('aあä', encoding='utf-8')
        fqfn = os.path.join(self.media_dir, 'test.txt')
        open(fqfn, 'wb').write(image_content.encode('utf-8'))
        
        expected_hash = '1f40fc92da241694750979ee6cf582f2d5d7d28e18335de05abc54d0560e0f5302860c652bf08d560252aa5e74210546f369fbbbce8c12cfc7957b2652fe9a75' # pylint: disable=C0301
        h = utils.get_text_hash(image_content)
        self.assertEqual(h, expected_hash)
        h = utils.get_file_hash(fqfn)
        self.assertEqual(h, expected_hash)
        h = utils.get_text_hash(open(fqfn, 'rb').read())
        self.assertEqual(h, expected_hash)

    def test_reading_file(self):
        call_command('loaddata', 'test_files.json')
        self.assertEqual(File.objects.count(), 1)
        response = self.client.get('/files/1.txt')
        if hasattr(response, 'streaming_content'):
            content = list(response.streaming_content)[0]
        else:
            content = response.content
        self.assertEqual(content, b'1234567890')
        self.assertEqual(response['content-type'], 'text/plain')
        self.assertEqual(response['content-length'], '10')




import os
import hashlib

import six

from django.conf import settings

def is_fresh(name, content_hash):
    """
    Returns true if the file exists on the local filesystem and matches the
    content in the database. Returns false otherwise.
    """
    if not content_hash:
        return False
    
    # Check that the actual file exists.
    fqfn = os.path.join(settings.MEDIA_ROOT, name)
    fqfn = os.path.normpath(fqfn)
    if not os.path.isfile(fqfn):
        return False
    
    # Check for cached hash file.
    hash_fn = get_hash_fn(name)
    if os.path.isfile(hash_fn):
        return open(hash_fn).read().strip() == content_hash
    
    # Otherwise, calculate the hash of the local file.
    fqfn = os.path.join(settings.MEDIA_ROOT, name)
    fqfn = os.path.normpath(fqfn)
    if not os.path.isfile(fqfn):
        return False
    local_content_hash = get_file_hash(fqfn)
    return local_content_hash == content_hash

def get_hash_fn(name):
    """
    Returns the filename for the hash file.
    """
    fqfn = os.path.join(settings.MEDIA_ROOT, name)
    fqfn = os.path.normpath(fqfn)
    dirs, fn = os.path.split(fqfn)
    if not os.path.isdir(dirs):
        os.makedirs(dirs)
    fqfn_parts = os.path.split(fqfn)
    hash_fn = os.path.join(fqfn_parts[0],
        settings.DB_FILES_DEFAULT_HASH_FN_TEMPLATE % fqfn_parts[1])
    return hash_fn

def write_file(name, content, overwrite=False):
    """
    Writes the given content to the relative filename under the MEDIA_ROOT.
    """
    fqfn = os.path.join(settings.MEDIA_ROOT, name)
    fqfn = os.path.normpath(fqfn)
    if os.path.isfile(fqfn) and not overwrite:
        return
    dirs, fn = os.path.split(fqfn)
    if not os.path.isdir(dirs):
        os.makedirs(dirs)
    open(fqfn, 'wb').write(content)
    
    # Cache hash.
    hash_value = get_file_hash(fqfn)
    hash_fn = get_hash_fn(name)
    try:
        # Write out bytes in Python3.
        value = six.binary_type(hash_value, 'utf-8')
    except TypeError:
        value = hash_value
    open(hash_fn, 'wb').write(value)
    
    # Set ownership and permissions.
    uname = getattr(settings, 'DATABASE_FILES_USER', None)
    gname = getattr(settings, 'DATABASE_FILES_GROUP', None)
    if gname:
        gname = ':'+gname
    if uname:
        os.system('chown -RL %s%s "%s"' % (uname, gname, dirs))

    # Set permissions.
    perms = getattr(settings, 'DATABASE_FILES_PERMS', None)
    if perms:
        os.system('chmod -R %s "%s"' % (perms, dirs))

def get_file_hash(fin, force_encoding=None, encoding=None, errors=None, chunk_size=128):
    """
    Iteratively builds a file hash without loading the entire file into memory.
    """
    
    force_encoding = force_encoding or settings.DB_FILES_DEFAULT_ENFORCE_ENCODING
    
    encoding = encoding or settings.DB_FILES_DEFAULT_ENCODING
    
    errors = errors or settings.DB_FILES_DEFAULT_ERROR_METHOD
    
    if isinstance(fin, six.string_types):
        fin = open(fin, 'rb')
    h = hashlib.sha512()
    while 1:
        text = fin.read(chunk_size)
        if not text:
            break
        if force_encoding:
            if not isinstance(text, six.text_type):
                text = six.text_type(text, encoding=encoding, errors=errors)
            h.update(text.encode(encoding, errors))
        else:
            h.update(text)
    return h.hexdigest()

def get_text_hash_0004(text):
    """
    Returns the hash of the given text.
    """
    h = hashlib.sha512()
    if not isinstance(text, six.text_type):
        text = six.text_type(text, encoding='utf-8', errors='replace')
    h.update(text.encode('utf-8', 'replace'))
    return h.hexdigest()

def get_text_hash(text, force_encoding=None, encoding=None, errors=None):
    """
    Returns the hash of the given text.
    """
    
    force_encoding = force_encoding or settings.DB_FILES_DEFAULT_ENFORCE_ENCODING
    
    encoding = encoding or settings.DB_FILES_DEFAULT_ENCODING
    
    errors = errors or settings.DB_FILES_DEFAULT_ERROR_METHOD
        
    h = hashlib.sha512()
    if force_encoding:
        if not isinstance(text, six.text_type):
            text = six.text_type(text, encoding=encoding, errors=errors)
        h.update(text.encode(encoding, errors))
    else:
        h.update(text)
    return h.hexdigest()




try:
    # Removed in Django 1.6
    from django.conf.urls.defaults import url
except ImportError:
    from django.conf.urls import url
    
try:
    # Relocated in Django 1.6
    from django.conf.urls.defaults import pattern
except ImportError:
    # Completely removed in Django 1.10
    try:    
        from django.conf.urls import patterns
    except ImportError:
        patterns = None

import database_files.views

_patterns = [
#    url(r'^files/(?P<name>.+)$',
#        database_files.views.serve,
#        name='database_file'),
    url(r'^files/(?P<name>.+)$',
        database_files.views.serve_mixed,
        name='database_file'),
]

if patterns is None:
    urlpatterns = _patterns
else:
    urlpatterns = patterns('', *_patterns)



# encoding: utf-8
import datetime

from django.core.management import call_command
from django.db import models

from south.db import db
from south.v2 import DataMigration

class Migration(DataMigration):

    def forwards(self, orm):
        # Load any files referenced by existing models into the database.
        call_command('database_files_load')

    def backwards(self, orm):
        import database_files
        database_files.models.File.objects.all().delete()

    models = {
        'database_files.file': {
            'Meta': {'object_name': 'File'},
            '_content': ('django.db.models.fields.TextField', [], {'db_column': "'content'"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '255', 'db_index': 'True'}),
            'size': ('django.db.models.fields.PositiveIntegerField', [], {})
        }
    }

    complete_apps = ['database_files']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
from setuptools import setup, find_packages, Command

import database_files

CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))

def get_reqs(*fns):
    lst = []
    for fn in fns:
        for package in open(os.path.join(CURRENT_DIR, fn)).readlines():
            package = package.strip()
            if not package:
                continue
            lst.append(package.strip())
    return lst

try:
    long_description = read_md('README.md')
except:
    long_description = ''

setup(
    name='django-database-files-3000',
    version=database_files.__version__,
    description='A storage system for Django that stores uploaded files in both the database and file system.',
    long_description=long_description,
    author='Chris Spencer',
    author_email='chrisspen@gmail.com',
    url='http://github.com/chrisspen/django-database-files-3000',
    packages=find_packages(),
    #https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 6 - Mature',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    install_requires=get_reqs('pip-requirements-min-django.txt', 'pip-requirements.txt'),
    tests_require=get_reqs('pip-requirements-test.txt'),
)



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import DataMigration
from django.db import models
import base64

from database_files import utils

class Migration(DataMigration):

    def forwards(self, orm):
        "Write your forwards methods here."
        File = orm['database_files.File']
        q = File.objects.all()
        for f in q:
            f._content_hash = utils.get_text_hash_0004(base64.b64decode(f._content))
            f.save()

    def backwards(self, orm):
        "Write your backwards methods here."

    models = {
        'database_files.file': {
            'Meta': {'object_name': 'File'},
            '_content': ('django.db.models.fields.TextField', [], {'db_column': "'content'"}),
            '_content_hash': ('django.db.models.fields.CharField', [], {'max_length': '128', 'null': 'True', 'db_column': "'content_hash'", 'blank': 'True'}),
            'created_datetime': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now', 'db_index': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '255', 'db_index': 'True'}),
            'size': ('django.db.models.fields.PositiveIntegerField', [], {})
        }
    }

    complete_apps = ['database_files']
    symmetrical = True
